@@ -1,6 +1,6 @@
 package ClearCase::SyncTree;
 
-$VERSION = '0.56';
+$VERSION = '0.57';
 
 require 5.004;
 
@@ -97,53 +97,35 @@ sub clone_ct {
     return $ct;
 }
 
-sub protect {
-    my $self = shift;
-    $self->{ST_PROTECT} = shift if @_;
-    return $self->{ST_PROTECT};
+sub gen_accessors {
+    my @key = map {uc} @_;
+    no strict 'refs';
+    for (@key) {
+	my $var = "ST_$_";
+	my $meth = lc;
+	*$meth = sub {
+	    my $self = shift;
+	    $self->{$var} = shift if @_;
+	    return $self->{$var};
+	}
+    }
 }
-
-sub remove {
-    my $self = shift;
-    $self->{ST_REMOVE} = shift if @_;
-    return $self->{ST_REMOVE};
-}
-
-sub reuse {
-    my $self = shift;
-    $self->{ST_REUSE} = shift if @_;
-    return $self->{ST_REUSE};
-}
-
-sub vreuse {
-    my $self = shift;
-    $self->{ST_VREUSE} = shift if @_;
-    return $self->{ST_VREUSE};
-}
-
-sub ignore_co {
-    my $self = shift;
-    $self->{ST_IGNORE_CO} = shift if @_;
-    return $self->{ST_IGNORE_CO};
-}
-
-sub overwrite_co {
-    my $self = shift;
-    $self->{ST_OVERWRITE_CO} = shift if @_;
-    return $self->{ST_OVERWRITE_CO};
-}
-
-sub snapdest {
-    my $self = shift;
-    $self->{ST_SNAPDEST} = shift if @_;
-    return $self->{ST_SNAPDEST};
-}
-
-sub ctime {
-    my $self = shift;
-    $self->{ST_CTIME} = shift if @_;
-    return $self->{ST_CTIME};
+gen_accessors(qw(protect remove reuse vreuse lblver ignore_co overwrite_co
+		 snapdest ctime lbtype inclb cmp_func rellinks dstview));
+sub gen_flags {
+    my @key = map {uc} @_;
+    no strict 'refs';
+    for (@key) {
+	my $var = "ST_$_";
+	my $meth = lc;
+	*$meth = sub {
+	    my $self = shift;
+	    $self->{$var} = 1 if $_[0] || !defined(wantarray);
+	    return $self->{$var};
+	}
+    }
 }
+gen_flags(qw(label_mods no_cr no_cmp));
 
 sub comment {
     my $self = shift;
@@ -249,10 +231,23 @@ sub mvfsdrive {
     return $self->{ST_MVFSDRIVE};
 }
 
-sub dstview {
-    my $self = shift;
-    $self->{ST_DSTVIEW} = shift if @_;
-    return $self->{ST_DSTVIEW};
+sub ccsymlink {
+    my $dst = shift;
+    return 1 if -l $dst;
+    return 0 unless MSWIN || CYGWIN;
+    my $ct = new ClearCase::Argv({autochomp=>1, stderr=>0});
+    return $ct->des([qw(-fmt %m)], $dst)->qx eq 'symbolic link';
+}
+
+# readlink might work under some conditions (CC version, mount options, ...)
+sub readcclink {
+    my $dst = shift;
+    my $ret = readlink $dst;
+    return $ret if $ret || !(MSWIN || CYGWIN);
+    my $ct = new ClearCase::Argv({autochomp=>1});
+    $ret = $ct->ls($dst)->qx;
+    $ret =~ s%\\%/%g if MSWIN;
+    return (($ret =~ s/^.*? --> (.*)$/$1/)? $ret : '');
 }
 
 sub srcbase {
@@ -264,6 +259,18 @@ sub srcbase {
 	# File::Spec::Win32::rel2abs leaves trailing / on drive letter root.
 	$sbase =~ s%/*$%% if $sbase ne '/';
 	$self->{ST_SRCBASE} = $sbase;
+	*src_slink = sub { return -l shift };
+	*src_rlink = sub { return readlink shift };
+	if (MSWIN || CYGWIN) {
+	    my $ct = $self->clone_ct({autofail=>1, autochomp=>1});
+	    my $olddir = getcwd;
+	    $ct->_chdir($sbase) || die "$0: Error: $sbase: $!";
+	    if ($ct->pwv(['-s'])->qx !~ /\s+NONE\s+/) {
+		*src_slink = \&ccsymlink;
+		*src_rlink = \&readcclink;
+	    }
+	    $ct->_chdir($olddir);
+	}
     }
     return $self->{ST_SRCBASE};
 }
@@ -377,48 +384,6 @@ sub dstvob {
     return $self->{ST_DSTVOB};
 }
 
-sub lbtype {
-    my $self = shift;
-    $self->{ST_LBTYPE} = shift if @_;
-    return $self->{ST_LBTYPE};
-}
-
-sub inclb {
-    my $self = shift;
-    $self->{ST_INCLB} = shift if @_;
-    return $self->{ST_INCLB};
-}
-
-sub label_mods {
-    my $self = shift;
-    $self->{ST_LABEL_MODS} = 1 if $_[0] || !defined(wantarray);
-    return $self->{ST_LABEL_MODS};
-}
-
-sub no_cr {
-    my $self = shift;
-    $self->{ST_NO_CR} = 1 if $_[0] || !defined(wantarray);
-    return $self->{ST_NO_CR};
-}
-
-sub no_cmp {
-    my $self = shift;
-    $self->{ST_NO_CMP} = 1 if $_[0] || !defined(wantarray);
-    return $self->{ST_NO_CMP} || 0;
-}
-
-sub cmp_func {
-    my $self = shift;
-    $self->{ST_CMP_FUNC} = shift if @_;
-    return $self->{ST_CMP_FUNC};
-}
-
-sub rellinks {
-    my $self = shift;
-    $self->{ST_RELLINKS} = shift if @_;
-    return $self->{ST_RELLINKS};
-}
-
 sub srclist {
     my $self = shift;
     my $type = ref($_[0]) ? ${shift @_} : 'NORMAL';
@@ -510,9 +475,14 @@ sub dstcheck {
 sub vtcomp {
     my($self, $src, $dst) = @_;
     my $cmp = $self->cmp_func;
+    my $lb = $self->lblver;
+    if ($lb) {
+        my $lblver = "$dst\@\@/$lb";
+	$dst = $lblver if -r $lblver;
+    }
     return 0 unless $cmp->($src, $dst);
     my $vt = ClearCase::Argv->lsvtree([qw(-a -s -nco)]);
-    my @vt = grep {m%[\\/]\d*$%} $vt->args($dst)->qx;
+    my @vt = reverse grep {m%[\\/]\d*$%} $vt->args($dst)->qx;
     chomp @vt;
     my $sz = -s $src;
     for (@vt) {
@@ -525,33 +495,14 @@ sub vtcomp {
     return 1;
 }
 
-sub ccsymlink {
-    my $dst = shift;
-    return 1 if -l $dst;
-    return 0 unless MSWIN || CYGWIN;
-    my $ct = new ClearCase::Argv({autochomp=>1, stderr=>0});
-    return $ct->des([qw(-fmt %m)], $dst)->qx eq 'symbolic link';
-}
-
-# readlink might work under some conditions (CC version, mount options, ...)
-sub readcclink {
-    my $dst = shift;
-    my $ret = readlink $dst;
-    return $ret if $ret || !(MSWIN || CYGWIN);
-    my $ct = new ClearCase::Argv({autochomp=>1});
-    $ret = $ct->ls($dst)->qx;
-    $ret =~ s%\\%/%g if MSWIN;
-    return (($ret =~ s/^.*? --> (.*)$/$1/)? $ret : '');
-}
-
 sub _needs_update {
     my($self, $src, $dst, $comparator) = @_;
     my $update = 0;
-    if (-l $src && ccsymlink($dst)) {
-	my $srctext = readlink $src;
+    if (src_slink($src) && ccsymlink($dst)) {
+	my $srctext = src_rlink($src);
 	my $desttext = readcclink $dst;
 	$update = !defined($comparator) || ($srctext ne $desttext);
-    } elsif (! -l $src && ! ccsymlink($dst)) {
+    } elsif (! src_slink($src) && ! ccsymlink($dst)) {
 	if (!defined($comparator)) {
 	    $update = 1;
 	} elsif ($self->vreuse) {
@@ -561,7 +512,7 @@ sub _needs_update {
 	} else {
 	    $update = &$comparator($src, $dst);
 	}
-	die "$0: Error: failed comparing $src vs $dst: $!" if $update < 0;
+	$self->failm("failed comparing $src vs $dst: $!") if $update < 0;
     } else {
 	$update = 1;
     }
@@ -597,13 +548,36 @@ sub analyze {
     # comparing src/dst files.
     delete $self->{ST_ADD};
     delete $self->{ST_MOD};
+    my @sl = $dbase eq $self->{ST_MKBASE}? sort grep{-d $_}
+      $self->clone_ct->find($dbase, qw(-type l -print))->qx : ();
+    map { $_ = "/$_" } @sl if CYGWIN; # mismatch between conventions
+    if (@sl) {
+	my %sl = map{ $_ => 1} @sl;
+	for my $l (@sl) {
+	    my $s = $l;
+	    $s =~ s%^\Q$dbase\E/(.*)$%$1%;
+	    if (exists $self->{ST_SRCMAP}->{$s}) {
+		$s = join('/', $sbase, $s);
+		delete $sl{$l} if src_slink($s);
+	    }
+	}
+	@sl = sort keys %sl;
+    }
     my $comparator = $self->no_cmp ? undef : $self->cmp_func;
-    for (sort keys %{$self->{ST_SRCMAP}}) {
+    SRC: for (sort keys %{$self->{ST_SRCMAP}}) {
 	next if $self->{ST_SRCMAP}->{$_}->{type} &&
 		$self->{ST_SRCMAP}->{$_}->{type} !~ /$type/;
 	my $src = join('/', $sbase, $_);
-	$src = $_ if ! -e $src && (MSWIN || ! -l $src);
+	$src = $_ unless -e $src || src_slink($src);
 	my $dst = join('/', $dbase, $self->{ST_SRCMAP}->{$_}->{dst} || $_);
+	for my $s (@sl) {
+	    if ($dst =~ /^\Q$s\E/) {
+		$self->{ST_DIRLNK}->{$s} = 1;
+		$self->{ST_ADD}->{$_}->{src} = $src;
+		$self->{ST_ADD}->{$_}->{dst} = $dst;
+		next SRC;
+	    }
+	}
 	# It's possible for a symlink to not satisfy -e if it's dangling.
 	# Case-insensitive file test operators are a problem on Windows.
 	# You cannot modify files when they don't exist under the proper name.
@@ -617,6 +591,20 @@ sub analyze {
 	    }
 	}
     }
+    if ($self->{ST_DIRLNK}) {
+	my @rem;
+	my @slst = sort keys %{$self->{ST_DIRLNK}};
+	for (reverse @slst) {
+	    for my $l (@slst) {
+		if (/^\Q$l\E./) {
+		    push @rem, $_;
+		    last;
+		}
+	    }
+	}
+	delete @{$self->{ST_DIRLNK}}{@rem} if @rem;
+	unlink $self->{ST_DIRLNK} unless keys %{$self->{ST_DIRLNK}};
+    }
     # Last, check for subtractions but only if asked - it's potentially
     # expensive and error-prone.
     return unless $self->remove;
@@ -630,7 +618,9 @@ sub analyze {
 	}
 	# Get a relative path from the absolute path.
 	(my $relpath = $path) =~ s%^\Q$dbase\E\W?%%;
-	if (-d $path) {
+	if (ccsymlink($path)) { # Vob link
+	    $files{$relpath} = $path;
+	} elsif (-d $path) {
 	    $dirs{$path} = $relpath;
 	} elsif (-f $path) {
 	    $files{$relpath} = $path;
@@ -654,6 +644,13 @@ sub preview {
     my $self = shift;
     my $indent = ' ' x 4;
     my($adds, $mods, $subs) = (0, 0, 0);
+    if ($self->{ST_DIRLNK}) {
+	my $dl = keys %{$self->{ST_DIRLNK}};
+	print "Removing $dl directory symlinks:\n";
+	for (sort keys %{$self->{ST_DIRLNK}}) {
+	    print "${indent}$_\n";
+	}
+    }
     if ($self->{ST_ADD}) {
 	$adds = keys %{$self->{ST_ADD}};
 	print "Adding $adds elements:\n";
@@ -739,9 +736,20 @@ sub branchco {
     return $rc;
 }
 
+sub rmdirlinks {
+    my $self = shift;
+    return unless $self->{ST_DIRLNK};
+    my $lsco = ClearCase::Argv->lsco([qw(-s -d -cview)]);
+    for (sort {$b cmp $a} keys %{$self->{ST_DIRLNK}}) {
+	my $dad = dirname $_;
+	$self->branchco(1, $dad) unless $lsco->args($dad)->qx;
+	$self->clone_ct->rm($_)->system;
+    }
+}
+
 sub mkrellink {
     my ($self, $src) = @_;
-    my $txt = readlink($src);
+    my $txt = src_rlink($src);
     my $sbase = $self->srcbase;
     return $txt unless $self->{ST_RELLINKS} and ($txt =~ /^$sbase/);
     $txt =~ s%^$sbase/(.*)%$1%;
@@ -773,13 +781,31 @@ sub skimdir {
     }
 }
 
+sub vtree {
+    my ($self, $dir) = @_;
+    if (!exists $self->{ST_VT}->{$dir}) {
+	my $vt = ClearCase::Argv->lsvtree({autochomp=>1}, [qw(-a -s -nco)]);
+	# optimization: branch/0 of a directory is either empty or duplicate
+	my @vt = reverse grep { m%[/\\](\d+)$% && $1>=1 } $vt->args($dir)->qx;
+	$self->{ST_VT}->{$dir} = \@vt;
+    }
+    return $self->{ST_VT}->{$dir};
+}
+
+# Once a directory version was found, move it first in the list for next tries
+sub raise_dver {
+    my ($self, $i, $dir) = @_;
+    return unless $i;
+    my $vt = $self->{ST_VT}->{$dir};
+    my $ver = splice @{$vt}, $i, 1;
+    unshift @{$vt}, $ver;
+}
+
 # Reuse from removed elements, or create as view private, directories
 sub reusemkdir {
     my ($self, $dref, $rref) = @_;
-    my (%found, $reused, %dfound);
+    my (%found, %dfound, %priv);
     my $snapview = $self->snapdest;
-    my $vt = ClearCase::Argv->lsvtree({autochomp=>1, stderr=>0},
-				                             [qw(-a -s -nco)]);
     my $ds = ClearCase::Argv->desc({stderr=>1},[qw(-s)]);
     my $dm = ClearCase::Argv->desc([qw(-fmt %m)]);
     my $rm = ClearCase::Argv->rm;
@@ -787,75 +813,84 @@ sub reusemkdir {
     my $ln = ClearCase::Argv->ln;
     for my $dst (sort keys %{$dref}) {
 	next if $dfound{$dst};
+	my $reused;
 	my($name, $dir) = fileparse($dst);
-	if ($rref->{$dst}) {
+	if (!$priv{$dir}) {
+	  if ($rref->{$dst}) {
 	    $self->branchco(1, $dir) unless $lsco->args($dir)->qx;
 	    $rm->args($dst)->system;
-	}
-	my @vtree = reverse grep { m%[/\\]\d+$% } $vt->args($dir)->qx;
-      VER: for (@vtree) {
+	  }
+	  my $i = -1; #index in the vtree list
+	VER: for (@{$self->vtree($dir)}) {
+	    $i++;
 	    my $dirext = "$_/$name";
 	    # case-insensitive file test operator on Windows is a problem
 	    if ($snapview ? $ds->args($dirext)->qx !~ /Error:/ : ecs($dirext)) {
-		next if $dm->args($dirext)->qx eq 'file element';
-		while (ccsymlink($dirext)) {
-		    $name = readcclink $dirext;
-		    $name =~ s/\@\@$//;
-		    # consider only relative, and local symlinks
-		    next VER if !ecs($dirext) ||
-		                       $dm->args($dirext)->qx eq 'file element';
+	      next if $dm->args($dirext)->qx eq 'file element';
+	      while (ccsymlink($dirext)) {
+		$name = readcclink $dirext;
+		$name =~ s/\@\@$//;
+		$dirext = "$_/$name";
+		# consider only relative, and local symlinks
+		next VER if !ecs($dirext) ||
+		  $dm->args($dirext)->qx eq 'file element';
+	      }
+	      $reused = 1;
+	      $self->raise_dver($i, $dir);
+	      $self->branchco(1, $dir) unless $lsco->args($dir)->qx;
+	      $ln->args($dirext, $dst)->system;
+	      # Need to reevaluate all the files under this dir
+	      # The case of implicit dirs, is recorded as '.'
+	      my $d = $dref->{$dst} eq '.'? '' : $dref->{$dst} . '/';
+	      $self->skimdir($dst, $d) if $self->remove;
+	      my $cmp = $self->no_cmp ? undef : $self->cmp_func;
+	      my @keys = sort $d? grep m%^\Q$d\E%, keys %{$self->{ST_ADD}}
+		: keys %{$self->{ST_ADD}};
+	      for my $e (@keys) {
+		my $edst = join '/', $self->dstbase, $e;
+		my @intdir = split m%/%, $e;
+		pop @intdir;
+		if (@intdir) {
+		  my $dd = $self->dstbase;
+		  my $pf = '';
+		  while (my $id = shift @intdir) {
+		    $dd = join '/', $dd, $id;
+		    $pf = $pf . $id . '/';
+		    $self->skimdir($dd, $pf) unless $dfound{$dd}++;
+		  }
 		}
-		$reused = 1;
-		$self->branchco(1, $dir) unless $lsco->args($dir)->qx;
-		$ln->args($dirext, $dst)->system;
-	        # Need to reevaluate all the files under this dir
-	        # The case of implicit dirs, is recorded as '.'
-		my $d = $dref->{$dst} eq '.'? '' : $dref->{$dst} . '/';
-		$self->skimdir($dst, $d) if $self->remove;
-		my $cmp = $self->no_cmp ? undef : $self->cmp_func;
-		my @keys = sort $d? grep m%^\Q$d\E%, keys %{$self->{ST_ADD}}
-		                                      : keys %{$self->{ST_ADD}};
-		for my $e (@keys) {
-		    my $edst = join '/', $self->dstbase, $e;
-		    my @intdir = split m%/%, $e;
-		    pop @intdir;
-		    if (@intdir) {
-			my $dd = $self->dstbase;
-			my $pf = '';
-			while (my $id = shift @intdir) {
-			    $dd = join '/', $dd, $id;
-			    $pf = $pf . $id . '/';
-			    $self->skimdir($dd, $pf) unless $dfound{$dd}++;
-			}
-		    }
-		    # Problem: does it match the type under srcbase?
-		    if (-d $edst) { # We know it ought to be empty
-			opendir(DIR, $edst);
-			my @f = grep !m%^\.\.?$%, readdir DIR;
-			closedir DIR;
-			if (@f) {
-			    $self->branchco(1, $edst)
-			                          unless $lsco->args($edst)->qx;
-			    $rm->args(map{join '/', $edst, $_} @f)->system;
-			}
-			$dfound{$edst}++; #Skip in this loop
-			next;
-		    }
-		    if (exists($self->{ST_ADD}->{$e}->{dst})) {
-			my $src = $self->{ST_ADD}->{$e}->{src};
-			my $dst = $self->{ST_ADD}->{$e}->{dst};
-			if (-e $dst) {
-			    $self->{ST_MOD}->{$e} = $self->{ST_ADD}->{$e}
-			              if $self->_needs_update($src, $dst, $cmp);
-			    $found{$e}++; #Remove from the add list
-			}
-		    }
+		# Problem: does it match the type under srcbase?
+		if (-d $edst and !ccsymlink($edst)) { # We know it is empty
+		  opendir(DIR, $edst);
+		  my @f = grep !m%^\.\.?$%, readdir DIR;
+		  closedir DIR;
+		  if (@f) {
+		    $self->branchco(1, $edst)
+		      unless $lsco->args($edst)->qx;
+		    $rm->args(map{join '/', $edst, $_} @f)->system;
+		  }
+		  $dfound{$edst}++; #Skip in this loop
+		  next;
 		}
-		last;
+		if (exists($self->{ST_ADD}->{$e}->{dst})) {
+		  my $src = $self->{ST_ADD}->{$e}->{src};
+		  my $dst = $self->{ST_ADD}->{$e}->{dst};
+		  if (-e $dst) {
+		    $self->{ST_MOD}->{$e} = $self->{ST_ADD}->{$e}
+		      if $self->_needs_update($src, $dst, $cmp);
+		    $found{$e}++; #Remove from the add list
+		  }
+		}
+	      }
+	      last;
 	    }
+	  }
 	}
 	if (!$reused) {
-	    mkpath($dst, 0, 0777) || die "$0: Error: $dst: $!";
+	    my $err;
+	    mkpath($dst, {error => \$err, verbose => 0, mode => 0777});
+	    $self->failm(join(': ', %{$err->[0]})) if $err and @{$err};
+	    $priv{"${dst}/"}++;
 	}
     }
     return %found;
@@ -883,8 +918,9 @@ sub add {
 	# in the way if added in _mkbase as view private; ignore failures
 	rmdir($_) for reverse sort @{$self->{ST_IMPLICIT_DIRS}};
 	for my $d (sort keys %{$self->{ST_ADD}}) {
+	    my $src = $self->{ST_ADD}->{$d}->{src};
 	    my $dst = $self->{ST_ADD}->{$d}->{dst};
-	    $dir{$dst} = $d if -d $d; # only empty dirs are explicitly added
+	    $dir{$dst} = $d if -d $src && !src_slink($src); # empty dir
 	    $self->recadd($d, $dst, \%dir, \%rm, \%dseen);
 	}
 	my %found = $self->reusemkdir(\%dir, \%rm);
@@ -893,28 +929,32 @@ sub add {
     for (sort keys %{$self->{ST_ADD}}) {
 	my $src = $self->{ST_ADD}->{$_}->{src};
 	my $dst = $self->{ST_ADD}->{$_}->{dst};
-	if (-d $src && ! -l $src) { # Already checked in the reuse case
-	    -e $dst || mkpath($dst, 0, 0777) || die "$0: Error: $dst: $!";
+	my $err;
+	if (-d $src && ! src_slink($src)) { # Already checked in the reuse case
+	    -e $dst || mkpath($dst, {error=>\$err, verbose=>0, mode=>0777});
+	    $self->failm(join(': ', %{$err->[0]})) if $err and @{$err};
 	} elsif (-e $src) {
 	    my $dad = dirname($dst);
-	    -d $dad || mkpath($dad, 0, 0777) || die "$0: Error: $dad: $!";
-	    if (-l $src) {
-		open(SLINK, ">$dst$lext") || die "$0: Error: $dst$lext: $!";
+	    -d $dad || mkpath($dad, {error=>\$err, verbose=>0, mode=>0777});
+	    $self->failm(join(': ', %{$err->[0]})) if $err and @{$err};
+	    if (src_slink($src)) {
+		open(SLINK, ">$dst$lext") || $self->failm("$dst$lext: $!");
 		print SLINK $self->mkrellink($src), "\n";;
 		close(SLINK);
 	    } else {
-		copy($src, $dst) || die "$0: Error: $_: $!";
-		utime(time(), (stat $src)[9], $dst) ||
-			warn "Warning: $dst: touch failed";
 		$self->{ST_CI_FROM}->{$_} = $self->{ST_ADD}->{$_}
-					    if !exists($self->{ST_PRE}->{$dst});
+					   if !exists($self->{ST_PRE}->{$dst});
 	    }
+	} elsif (src_slink($src)) { #Dangling symlink: import
+	    open(SLINK, ">$dst$lext") || $self->failm("$dst$lext: $!");
+	    print SLINK $self->mkrellink($src), "\n";;
+	    close(SLINK);
 	} else {
-	    warn "$0: Error: $src: no such file or directory\n";
-	    $ct->fail;
+	    $ct->failm("$src: no such file or directory");
 	}
     }
-    my @candidates = sort $self->_lsprivate(1);
+    my @candidates = sort ($self->_lsprivate(1),
+			   map { $_->{dst} } values %{$self->{ST_CI_FROM}});
     return if !@candidates;
     # We'll be separating the elements-to-be into files and directories.
     my(%files, @symlinks, %dirs);
@@ -933,8 +973,8 @@ sub add {
     }
     $self->branchco(1, keys %dirs) if keys %dirs;
     # Process candidate directories here, then do files below.
-    my $mkdir = $self->clone_ct->argv({autofail=>0, autochomp=>0},
-				                      'mkdir', $self->comment);
+    my $mkdir = $self->clone_ct->mkdir({autofail=>0, autochomp=>0},
+				                               $self->comment);
     for my $cand (@candidates) {
 	if (! -d $cand) {
 	    if ($cand =~ /$lext$/) {
@@ -967,7 +1007,7 @@ sub add {
 	}
 	while (defined(my $i = readdir(DIR))) {
 	    next if $i eq '.' || $i eq '..';
-	    rename("$tmpdir/$i", "$cand/$i") || die "$0: Error: $cand/$i: $!";
+	    rename("$tmpdir/$i", "$cand/$i") || $self->failm("$cand/$i: $!");
 	}
 	closedir DIR;
 	rmdir $tmpdir || warn "$0: Error: $tmpdir: $!";
@@ -976,31 +1016,32 @@ sub add {
     # Optionally, reconstitute an old element of the same name if present.
     if ($self->reuse) {
 	my $snapview = $self->snapdest;
-	my $vt = ClearCase::Argv->lsvtree([qw(-a -s -nco)]);
 	my $ds = ClearCase::Argv->desc({stderr=>1}, [qw(-s)]);
 	my $dm = ClearCase::Argv->desc([qw(-fmt %m)]);
 	my $ln = ClearCase::Argv->ln;
 	my %reused;
 	for my $elem (keys %files) {
 	    my($name, $dir) = fileparse($elem);
-	    chomp(my @vtree = reverse $vt->args($dir)->qx);
-	    for (@vtree) {
-		next unless m%[/\\](\d+)$% && $1 > 0;	# optimization
+	    my $i = -1;
+	  DVER:
+	    for (@{$self->vtree($dir)}) {
+		$i++;
 		my $dirext = "$_/$name@@";
 		# case-insensitive file test operator on Windows is a problem
 		if ($snapview ? $ds->args($dirext)->qx !~ /Error:/ :
 							    ecs("$_/$name")) {
-		    next if $dm->args("$_/$name")->qx eq 'directory element';
+		    next if $dm->args("$_/$name")->qx =~ /^directory /;
 		    while (ccsymlink("$_/$name")) {
 		        $name = readcclink "$_/$name";
 			$name =~ s/\@\@$//;
-			next if !ecs("$_/$name") ||
-			     $dm->args("$_/$name")->qx eq 'directory element';
+			next DVER if !ecs("$_/$name") ||
+			            $dm->args("$_/$name")->qx =~ /^directory /;
 		    }
 		    $reused{$elem} = 1;
 		    delete $files{$elem};
 		    unlink($elem);
 		    $ln->args("$_/$name", $elem)->system;
+		    $self->raise_dver($i, $dir);
 		    last;
 		}
 	    }
@@ -1036,7 +1077,15 @@ sub add {
 	    }
 	}
     }
-
+    for (sort keys %{$self->{ST_CI_FROM}}) {
+	my $dst = $self->{ST_CI_FROM}->{$_}->{dst};
+	if ($files{$dst}) {
+	    my $src = $self->{ST_CI_FROM}->{$_}->{src};
+	    copy($src, $dst) || $ct->failm("$_: $!");
+	    utime(time(), (stat $src)[9], $dst) ||
+	                                    warn "Warning: $dst: touch failed";
+	}
+    }
     # Now do the files in one fell swoop.
     $ct->mkelem($self->comment, sort keys %files)->system if %files;
 
@@ -1050,7 +1099,7 @@ sub add {
 	chomp(my $txt = <SLINK>);
 	close SLINK;
 	unlink $symlink;
-	$ct->argv('ln', ['-s'], $txt, $lnk)->system;
+	$ct->ln(['-s'], $txt, $lnk)->system;
     }
 }
 
@@ -1073,7 +1122,7 @@ sub modify {
     return if !keys %{$self->{ST_MOD}};
     my(%files, %symlinks);
     for (keys %{$self->{ST_MOD}}) {
-	if (-l $self->{ST_MOD}->{$_}->{src}) {
+	if (src_slink($self->{ST_MOD}->{$_}->{src})) {
 	    $symlinks{$_}++;
 	} else {
 	    $files{$_}++;
@@ -1089,6 +1138,7 @@ sub modify {
 	for my $key (sort keys %files) {
 	    my $src = $self->{ST_MOD}->{$key}->{src};
 	    my $dst = $self->{ST_MOD}->{$key}->{dst};
+	    my $new;
 	    if (ccsymlink($dst)) {
 	        # The source is a file, but the destination is a symlink: look
 	        # (recursively) at what this one points to, and link in this
@@ -1114,19 +1164,14 @@ sub modify {
 		$self->branchco(1, $dir) unless $lsco->args($dir)->qx;
 		$self->clone_ct->rm($dst)->system; #remove the first symlink
 		if ($dangling) {
-		    if ($self->no_cr) {
-			if (!copy($src, $dst)) {
-			    warn "$0: Error: $dst: $!\n";
-			    $rm->fail;
-			}
-			utime(time(), (stat $src)[9], $dst) ||
-			                    warn "Warning: $dst: touch failed";
-			$self->clone_ct->mkelem($self->comment, $dst)->system;
-		    } else {
-			my @ptime = qw(-pti) unless $self->ctime;
-			$self->clone_ct->ci([@ptime, @{$self->comment},
-				      qw(-ide -rm -from), $src], $dst)->system;
+		    if (!copy($src, $dst)) {
+			warn "$0: Error: $dst: $!\n";
+			$rm->fail;
 		    }
+		    utime(time(), (stat $src)[9], $dst) ||
+		      warn "Warning: $dst: touch failed";
+		    $self->clone_ct->mkelem($self->comment, $dst)->system;
+		    $new = 1;
 		    delete $self->{ST_MOD}->{$key};
 		    push @del, $key;
 		} else {
@@ -1147,22 +1192,25 @@ sub modify {
 		    }
 		}
 	    }
-	    push(@toco, $dst) if !exists($self->{ST_PRE}->{$dst});
+	    push(@toco, $dst) unless exists($self->{ST_PRE}->{$dst}) || $new;
 	}
 	$self->branchco(0, @toco) if @toco;
 	delete @files{@del};
 	for (sort keys %files) {
 	    my $src = $self->{ST_MOD}->{$_}->{src};
 	    my $dst = $self->{ST_MOD}->{$_}->{dst};
-	    if (!copy($src, $dst)) {
-		warn "$0: Error: $dst: $!\n";
-		$rm->fail;
-		next;
+	    next if exists($self->{ST_PRE}->{$dst});
+	    if ($self->no_cr) {
+		if (!copy($src, $dst)) {
+		    warn "$0: Error: $dst: $!\n";
+		    $rm->fail;
+		    next;
+		}
+		utime(time(), (stat $src)[9], $dst) ||
+				            warn "Warning: $dst: touch failed";
+	    } else {
+		$self->{ST_CI_FROM}->{$_} = $self->{ST_MOD}->{$_};
 	    }
-	    utime(time(), (stat $src)[9], $dst) ||
-				    warn "Warning: $dst: touch failed";
-	    $self->{ST_CI_FROM}->{$_} = $self->{ST_MOD}->{$_}
-			if !$self->no_cr && !exists($self->{ST_PRE}->{$dst});
 	}
     }
     if (keys %symlinks) {
@@ -1206,7 +1254,6 @@ sub subtract {
 		closedir(DIR);
 		next if @entries > 2;
 		push(@exdirs, $_);
-		delete $dirs{$_};
 	    }
 	}
 	last if !@exdirs;
@@ -1243,12 +1290,22 @@ sub label {
 	my @mods = $self->_lsco;
 	push @mods, @{$self->{ST_LBL}} if $self->{ST_LBL};
 	if (@mods) {
-	    $ctbool->mklabel([qw(-nc), $self->inclb], @mods)->system
+	    $ctbool->mklabel([qw(-nc -rep), $self->inclb], @mods)->system
 	                                                      if $self->inclb;
 	    $ctbool->mklabel([qw(-nc -rep), $lbtype], @mods)->system;
 	}
     } else {
-	$ctbool->mklabel([qw(-nc -rep -rec), $lbtype], $dbase)->system;
+	my $lbl = $self->lblver;
+        if ($lbl) {
+	    my $ct = $self->clone_ct({autochomp=>1, autofail=>0, stderr=>0});
+	    my @rv = grep{ s/^(.*?)(?:@@(.*))/$1/ &&
+			     ($2 =~ /CHECKEDOUT$/ || !-r "$_\@\@/$lbl") }
+	      $ct->ls([qw(-r -vob -s)], $dbase)->qx,
+	      $ct->ls([qw(-d -vob -s)], $dbase)->qx;
+	    $ctbool->mklabel([qw(-nc -rep), $lbtype], $dbase, @rv)->system;
+        } else {
+	    $ctbool->mklabel([qw(-nc -rep -rec), $lbtype], $dbase)->system;
+	}
 	# Possibly move the label back to the right versions
 	$ctbool->mklabel([qw(-nc -rep), $lbtype], @{$self->{ST_LBL}})->system
 	                                                   if $self->{ST_LBL};
@@ -1320,6 +1377,7 @@ sub checkin {
 	    my $dst = $self->{ST_CI_FROM}->{$_}->{dst};
 	    $ct->ci([@ptime, @cmnt, qw(-ide -rm -from), $src], $dst)->system;
 	}
+	delete @{$self->{ST_MOD}}{keys %{$self->{ST_CI_FROM}}};
     }
     # Check-in first the files modified under the recorded names,
     # in case of hardlinks, since checking the other link first
@@ -1400,7 +1458,20 @@ sub cleanup {
     @todo = grep {!exists($self->{ST_PRE}->{$_})} @todo
 				    if $self->ignore_co || $self->overwrite_co;
     unshift(@todo, $dad) if $checkedout{$dad};
-    $ct->argv('unco', [qw(-rm)], sort {$b cmp $a} @todo)->system if @todo;
+    if ($self->{branchoffroot}) {
+	for (sort {$b cmp $a} @todo) {
+	    my $b = $ct->ls([qw(-s -d)], $_)->qx;
+	    $ct->unco([qw(-rm)], $_)->system;
+	    if ($b =~ s%^(.*)[\\/]CHECKEDOUT$%$1%) {
+		opendir BR, $b or next;
+		my @f = grep !/^(\.\.?|0|LATEST)$/, readdir BR;
+		closedir BR;
+		$ct->rmbranch([qw(-f)], $b)->system unless @f;
+	    }
+	}
+    } else {
+	$ct->unco([qw(-rm)], sort {$b cmp $a} @todo)->system if @todo;
+    }
 }
 
 # Undo current work and exit. May be called from an exception handler.
@@ -1412,6 +1483,12 @@ sub fail {
     exit(defined($rc) ? $rc : 2);
 }
 
+sub failm {
+    my ($self, $msg, $rc) = @_;
+    warn "$0: Error: $msg\n";
+    $self->fail($rc);
+}
+
 sub version {
     my $self = shift;
     return $ClearCase::SyncTree::VERSION;
@@ -1768,11 +1845,6 @@ The support for Cygwin normalizes the paths on the UNC syntax.
 
 =item *
 
-Subtraction of symlinks is currently unimplemented (it's just a little
-corner case I haven't gotten to).
-
-=item *
-
 If a file is removed via the -E<gt>subtract method and later added back
 via -E<gt>add, the result will be a new element (aka I<evil twin>).
 The -E<gt>reuse method (see) may be used to prevent evil twins.
